@@ -7,7 +7,7 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include "miscfuncs.h"
+#include "create_posix_thread.h"
 #include <unistd.h>
 #include <stdlib.h>
 #include "glob.h"
@@ -61,10 +61,10 @@ do_global_dtors ()
     }
 }
 
-static void __stdcall
+static void
 do_global_ctors (void (**in_pfunc)(), int force)
 {
-  if (!force && in_forkee)
+  if (!force && __in_forkee == FORKING)
     return;		// inherit constructed stuff from parent pid
 
   /* Run ctors backwards, so skip the first entry and find how many
@@ -84,7 +84,7 @@ do_global_ctors (void (**in_pfunc)(), int force)
  * A \@file is replaced with @file so that echo \@foo would print
  * @foo and not the contents of foo.
  */
-static bool __stdcall
+static bool
 insert_file (char *name, char *&cmd)
 {
   HANDLE f;
@@ -203,7 +203,7 @@ quoted (char *cmd, int winshell)
 			    && isalpha ((s)[2]) \
 			    && strchr ((s) + 3, '\\')))
 
-static int __stdcall
+static int
 globify (char *word, char **&argv, int &argc, int &argvlen)
 {
   if (*word != '~' && strpbrk (word, "?*[\"\'(){}") == NULL)
@@ -290,7 +290,7 @@ globify (char *word, char **&argv, int &argc, int &argvlen)
 
 /* Build argv, argc from string passed from Windows.  */
 
-static void __stdcall
+static void
 build_argv (char *cmd, char **&argv, int &argc, int winshell)
 {
   int argvlen = 0;
@@ -365,7 +365,7 @@ build_argv (char *cmd, char **&argv, int &argc, int winshell)
 }
 
 /* sanity and sync check */
-void __stdcall
+void
 check_sanity_and_sync (per_process *p)
 {
   /* Sanity check to make sure developers didn't change the per_process    */
@@ -374,14 +374,7 @@ check_sanity_and_sync (per_process *p)
   if (sizeof (per_process) != SIZEOF_PER_PROCESS)
     api_fatal ("per_process sanity check failed");
 
-  /* Make sure that the app and the dll are in sync. */
-
-  /* Complain if older than last incompatible change */
-  if (p->dll_major < CYGWIN_VERSION_DLL_EPOCH)
-    api_fatal ("cygwin DLL and APP are out of sync -- DLL version mismatch %u < %u",
-	       p->dll_major, CYGWIN_VERSION_DLL_EPOCH);
-
-  /* magic_biscuit != 0 if using the old style version numbering scheme.  */
+  /* magic_biscuit must be SIZEOF_PER_PROCESS.  */
   if (p->magic_biscuit != SIZEOF_PER_PROCESS)
     api_fatal ("Incompatible cygwin .dll -- incompatible per_process info %u != %u",
 	       p->magic_biscuit, SIZEOF_PER_PROCESS);
@@ -390,14 +383,6 @@ check_sanity_and_sync (per_process *p)
   if (p->api_major > cygwin_version.api_major)
     api_fatal ("cygwin DLL and APP are out of sync -- API version mismatch %u > %u",
 	       p->api_major, cygwin_version.api_major);
-
-#ifdef __i386__
-  /* This is a kludge to work around a version of _cygwin_common_crt0
-     which overwrote the cxx_malloc field with the local DLL copy.
-     Hilarity ensues if the DLL is not loaded while the process
-     is forking. */
-  __cygwin_user_data.cxx_malloc = &default_cygwin_cxx_malloc;
-#endif
 }
 
 child_info NO_COPY *child_proc_info;
@@ -542,15 +527,12 @@ get_cygwin_startup_info ()
     {
       if ((res->intro & OPROC_MAGIC_MASK) == OPROC_MAGIC_GENERIC)
 	multiple_cygwin_problem ("proc intro", res->intro, 0);
-      else if (res->cygheap != (void *) &_cygheap_start)
-	multiple_cygwin_problem ("cygheap base", (uintptr_t) res->cygheap,
-				 (uintptr_t) &_cygheap_start);
 
       unsigned should_be_cb = 0;
       switch (res->type)
 	{
 	  case _CH_FORK:
-	    in_forkee = true;
+	    __in_forkee = FORKING;
 	    should_be_cb = sizeof (child_info_fork);
 	    fallthrough;
 	  case _CH_SPAWN:
@@ -581,17 +563,10 @@ get_cygwin_startup_info ()
   return res;
 }
 
-#ifdef __x86_64__
 #define dll_data_start &__data_start__
 #define dll_data_end &__data_end__
 #define dll_bss_start &__bss_start__
 #define dll_bss_end &__bss_end__
-#else
-#define dll_data_start &_data_start__
-#define dll_data_end &_data_end__
-#define dll_bss_start &_bss_start__
-#define dll_bss_end &_bss_end__
-#endif
 
 void
 child_info_fork::handle_fork ()
@@ -645,7 +620,6 @@ bool
 child_info_spawn::get_parent_handle ()
 {
   parent = OpenProcess (PROCESS_VM_READ, FALSE, parent_winpid);
-  moreinfo->myself_pinfo = NULL;
   return !!parent;
 }
 
@@ -732,15 +706,6 @@ init_windows_system_directory ()
       RtlInitCountedUnicodeString (&windows_directory_path,
 	    windows_directory_buf,
 	    (windows_directory_length + 4) * sizeof (WCHAR));
-#ifdef __i386__
-      system_wow64_directory_length =
-	    GetSystemWow64DirectoryW (system_wow64_directory, MAX_PATH);
-      if (system_wow64_directory_length)
-	{
-	  system_wow64_directory[system_wow64_directory_length++] = L'\\';
-	  system_wow64_directory[system_wow64_directory_length] = L'\0';
-	}
-#endif /* __i386__ */
     }
 }
 
@@ -756,10 +721,6 @@ dll_crt0_0 ()
   SetErrorMode (SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX);
 
   lock_process::init ();
-  _impure_ptr = _GLOBAL_REENT;
-  _impure_ptr->_stdin = &_impure_ptr->__sf[0];
-  _impure_ptr->_stdout = &_impure_ptr->__sf[1];
-  _impure_ptr->_stderr = &_impure_ptr->__sf[2];
   user_data->impure_ptr = _impure_ptr;
   user_data->impure_ptr_ptr = &_impure_ptr;
 
@@ -770,8 +731,6 @@ dll_crt0_0 ()
   NtOpenProcessToken (NtCurrentProcess (), MAXIMUM_ALLOWED, &hProcToken);
   set_cygwin_privileges (hProcToken);
 
-  malloc_init_0 ();
-
   device::init ();
   do_global_ctors (&__CTOR_LIST__, 1);
   cygthread::init ();
@@ -806,10 +765,8 @@ dll_crt0_0 ()
   if (!dynamically_loaded)
     sigproc_init ();
 
-#ifdef __x86_64__
   /* See comment preceeding myfault_altstack_handler in exception.cc. */
   AddVectoredContinueHandler (0, myfault_altstack_handler);
-#endif
 
   debug_printf ("finished dll_crt0_0 initialization");
 }
@@ -818,7 +775,7 @@ static inline void
 main_thread_sinit ()
 {
   __sinit (_impure_ptr);
-  /* At this point, _impure_ptr == _global_impure_ptr == _GLOBAL_REENT is
+  /* At this point, _impure_ptr == _GLOBAL_REENT is
      initialized, but _REENT == _my_tls.local_clib doesn't know about it.
      It has been copied over from _GLOBAL_REENT in _cygtls::init_thread
      *before* the initialization took place.
@@ -826,15 +783,13 @@ main_thread_sinit ()
      As soon as the main thread calls a stdio function, this would be
      rectified.  But if another thread calls a stdio function on
      stdin/out/err before the main thread does, all the required
-     initialization of stdin/out/err will be done, but _REENT->__sdidinit
-     is *still* 0.  This in turn will result in a call to __sinit in the
+     initialization of stdin/out/err will be done, but _REENT->__cleanup
+     is *still* NULL.  This in turn will result in a call to __sinit in the
      wrong spot.  The input or output buffer will be NULLed and nothing is
      read or written in the first stdio function call in the main thread.
 
-     To fix this issue we have to copy over the relevant part of _GLOBAL_REENT
-     to _REENT here again. */
-  _REENT->__sdidinit = -1;
-  _REENT->__cleanup = _GLOBAL_REENT->__cleanup;
+     To fix this issue we set __cleanup to _cygtls::cleanup_early here. */
+  _REENT_CLEANUP(_REENT) = _cygtls::cleanup_early;
 }
 
 /* Take over from libc's crt0.o and start the application. Note the
@@ -848,19 +803,22 @@ dll_crt0_1 (void *)
 
   _my_tls.incyg++;
   /* Inherit "parent" exec'ed process sigmask */
-  if (spawn_info && !in_forkee)
-    _my_tls.sigmask = spawn_info->moreinfo->sigmask;
+  if (spawn_info && __in_forkee != FORKING)
+    _my_tls.sigmask = spawn_info->sigmask;
 
   if (dynamically_loaded)
     sigproc_init ();
 
+  /* Call this before accessing any files. */
+  RtlSetProcessPlaceholderCompatibilityMode (PHCM_EXPOSE_PLACEHOLDERS);
+
   check_sanity_and_sync (user_data);
 
   /* Initialize malloc and then call user_shared_initialize since it relies
      on a functioning malloc and it's possible that the user's program may
      have overridden malloc.  We only know about that at this stage,
      unfortunately. */
-  malloc_init_1 ();
+  malloc_init ();
   user_shared->initialize ();
 
 #ifdef CYGHEAP_DEBUG
@@ -885,7 +843,7 @@ dll_crt0_1 (void *)
 
   /* Initialize pthread mainthread when not forked and it is safe to call new,
      otherwise it is reinitalized in fixup_after_fork */
-  if (!in_forkee)
+  if (__in_forkee != FORKING)
     {
       pthread::init_mainthread ();
       _pei386_runtime_relocator (user_data);
@@ -895,12 +853,8 @@ dll_crt0_1 (void *)
   strace.microseconds ();
 #endif
 
-  /* Initialize debug muto, if DLL is built with --enable-debugging.
-     Need to do this before any helper threads start. */
-  debug_init ();
-
   cygbench ("pre-forkee");
-  if (in_forkee)
+  if (__in_forkee == FORKING)
     {
       /* Make sure to restore the TEB's stack info.  If guardsize is -1 the
 	 stack has been provided by the application and must not be deallocated
@@ -1025,22 +979,27 @@ dll_crt0_1 (void *)
     {
       /* Create a copy of Cygwin's version of __argv so that, if the user makes
 	 a change to an element of argv[] it does not affect Cygwin's argv.
-	 Changing the the contents of what argv[n] points to will still
-	 affect Cygwin.  This is similar (but not exactly like) Linux. */
-      char *newargv[__argc + 1];
-      char **nav = newargv;
-      char **oav = __argv;
-      while ((*nav++ = *oav++) != NULL)
-	continue;
+	 Changing the contents of what argv[n] points to will still affect
+	 Cygwin.  This is similar (but not exactly like) Linux.
+
+	 We used to allocate newargv on the stack, but all the rest of the
+	 argument and environment handling does not depend on the stack,
+	 as it does on Linux.  In fact, everything is stored by the parent
+	 in the cygheap, so the only reason this may fail is if we're out
+	 of resources in a big way.  If this malloc fails, we could either
+	 fail the entire process execution, or we could proceed with the
+	 original argv and potentially affect output of /proc/self/cmdline.
+	 We opt for the latter here because it's the lesser evil. */
+      char **newargv = (char **) malloc ((__argc + 1) * sizeof (char *));
+      if (newargv)
+	memcpy (newargv, __argv, (__argc + 1) * sizeof (char *));
+      else
+	newargv = __argv;
       /* Handle any signals which may have arrived */
       sig_dispatch_pending (false);
       _my_tls.call_signal_handler ();
       _my_tls.incyg--;	/* Not in Cygwin anymore */
-#ifdef __x86_64__
-      cygwin_exit (user_data->main (__argc, newargv, __cygwin_environ));
-#else
-      cygwin_exit (user_data->main (__argc, newargv, *user_data->envptr));
-#endif
+      cygwin_exit (user_data->main (__argc, newargv, environ));
     }
   __asm__ ("				\n\
 	.global _cygwin_exit_return	\n\
@@ -1051,10 +1010,9 @@ __cygwin_exit_return:			\n\
 ");
 }
 
-extern "C" void __stdcall
+extern "C" void
 _dll_crt0 ()
 {
-#ifdef __x86_64__
   /* Starting with Windows 10 rel 1511, the main stack of an application is
      not reproducible if a 64 bit process has been started from a 32 bit
      process.  Given that we have enough virtual address space on 64 bit
@@ -1063,18 +1021,15 @@ _dll_crt0 ()
      under our own control and avoids collision with the OS. */
   if (!dynamically_loaded)
     {
-      if (!in_forkee)
+      if (__in_forkee != FORKING)
 	{
 	  /* Must be static since it's referenced after the stack and frame
 	     pointer registers have been changed. */
 	  static PVOID allocationbase;
-	  SIZE_T commitsize = in_forkee ? (PBYTE) fork_info->stackbase
-					  - (PBYTE) fork_info->stacklimit
-					: 0;
-	  PVOID stackaddr = create_new_main_thread_stack (allocationbase,
-							  commitsize);
+	  PVOID stackaddr = create_new_main_thread_stack (allocationbase);
 	  if (stackaddr)
 	    {
+#ifdef __x86_64__
 	      /* Set stack pointer to new address.  Set frame pointer to
 	         stack pointer and subtract 32 bytes for shadow space. */
 	      __asm__ ("\n\
@@ -1082,6 +1037,9 @@ _dll_crt0 ()
 		       movq  %%rsp, %%rbp  \n\
 		       subq  $32,%%rsp     \n"
 		       : : [ADDR] "r" (stackaddr));
+#else
+#error unimplemented for this target
+#endif
 	      /* We're on the new stack now.  Free up space taken by the former
 		 main thread stack and set DeallocationStack correctly. */
 	      VirtualFree (NtCurrentTeb ()->DeallocationStack, 0, MEM_RELEASE);
@@ -1091,11 +1049,6 @@ _dll_crt0 ()
       else
 	fork_info->alloc_stack ();
     }
-#else
-  main_environ = user_data->envptr;
-  if (in_forkee)
-    fork_info->alloc_stack ();
-#endif
 
   fesetenv (FE_DFL_ENV);
   _main_tls = &_my_tls;
@@ -1106,7 +1059,7 @@ void
 dll_crt0 (per_process *uptr)
 {
   /* Set the local copy of the pointer into the user space. */
-  if (!in_forkee && uptr && uptr != user_data)
+  if (__in_forkee != FORKING && uptr && uptr != user_data)
     {
       memcpy (user_data, uptr, per_process_overwrite);
       *(user_data->impure_ptr_ptr) = _GLOBAL_REENT;
@@ -1115,7 +1068,7 @@ dll_crt0 (per_process *uptr)
 }
 
 /* This must be called by anyone who uses LoadLibrary to load cygwin1.dll.
-   You must have CYGTLS_PADSIZE bytes reserved at the bottom of the stack
+   You must have __CYGTLS_PADSIZE__ bytes reserved at the bottom of the stack
    calling this function, and that storage must not be overwritten until you
    unload cygwin1.dll, as it is used for _my_tls.  It is best to load
    cygwin1.dll before spawning any additional threads in your process.
@@ -1125,18 +1078,10 @@ dll_crt0 (per_process *uptr)
 extern "C" void
 cygwin_dll_init ()
 {
-#ifdef __i386__
-  static char **envp;
-#endif
   static int _fmode;
 
   user_data->magic_biscuit = sizeof (per_process);
-
-#ifdef __i386__
-  user_data->envptr = &envp;
-#endif
   user_data->fmode_ptr = &_fmode;
-
   _dll_crt0 ();
 }
 
@@ -1158,7 +1103,7 @@ __main (void)
   sig_dispatch_pending (true);
 }
 
-void __reg1
+void
 do_exit (int status)
 {
   syscall_printf ("do_exit (%d), exit_state %d", status, exit_state);
@@ -1260,7 +1205,6 @@ cygwin_atexit (void (*fn) (void))
   int res;
 
   dll *d = dlls.find ((void *) _my_tls.retaddr ());
-#ifdef __x86_64__
   /* x86_64 DLLs created with GCC 4.8.3-3 register __gcc_deregister_frame
      as atexit function using a call to atexit, rather than __cxa_atexit.
      Due to GCC's tail call optimizing, cygwin_atexit doesn't get the correct
@@ -1282,7 +1226,6 @@ cygwin_atexit (void (*fn) (void))
      use the statically linked atexit function though, as outlined above. */
   if (!d)
     d = dlls.find ((void *) fn);
-#endif
   res = d ? __cxa_atexit ((void (*) (void *)) fn, NULL, d->handle) : atexit (fn);
   return res;
 }
@@ -1306,15 +1249,13 @@ extern "C" void
 vapi_fatal (const char *fmt, va_list ap)
 {
   char buf[4096];
-  int n = __small_sprintf (buf, "%P: *** fatal error %s- ", in_forkee ? "in forked process " : "");
+  int n = __small_sprintf (buf, "%P: *** fatal error %s- ",
+				(__in_forkee == FORKING)
+				? "in forked process " : "");
   __small_vsprintf (buf + n, fmt, ap);
   va_end (ap);
   strace.prntf (_STRACE_SYSTEM, NULL, "%s", buf);
-
-#ifdef DEBUGGING
-  try_to_debug ();
-#endif
-  cygwin_stackdump ();
+  api_fatal_debug();
   myself.exit (__api_fatal_exit_val);
 }
 
@@ -1330,7 +1271,7 @@ api_fatal (const char *fmt, ...)
 void
 multiple_cygwin_problem (const char *what, uintptr_t magic_version, uintptr_t version)
 {
-  if (_cygwin_testing && (strstr (what, "proc") || strstr (what, "cygheap")))
+  if (_cygwin_testing && (strstr (what, "proc")))
     {
       child_proc_info->type = _CH_WHOOPS;
       return;
@@ -1353,7 +1294,7 @@ are unable to find another cygwin DLL.",
 }
 
 #ifdef DEBUGGING
-void __reg1
+void
 cygbench (const char *s)
 {
   if (GetEnvironmentVariableA ("CYGWIN_BENCH", NULL, 0))
